/*
 * dvgrab.cc -- grabbing DV data and saving into an AVI file
 * Copyright (C) 2000 - 2002 Arne Schirmacher <arne@schirmacher.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Tag: $Name:  $
 *
 * Change log:
 * 
 * $Log: dvgrab.cc,v $
 * Revision 1.3  2003/02/23 02:02:12  ddennedy
 * added raw to stdout, updated docs
 *
 * Revision 1.2  2003/02/07 04:25:15  ddennedy
 * bugfix processing jpeg options
 *
 * Revision 1.1.1.1  2003/02/07 04:08:37  ddennedy
 * initial import of dvgrab
 *
 * Revision 1.4  2002/03/25 21:34:25  arne
 * Support for large (64 bit) files mostly completed
 *
 * Revision 1.3  2002/03/10 21:28:29  arne
 * release 1.1b1, 64 bit support for type 1 avis
 *
 * Revision 1.2  2002/03/04 19:22:43  arne
 * updated to latest Kino avi code
 *
 * Revision 1.1.1.1  2002/03/03 19:08:08  arne
 * import of version 1.01
 *
 */

/** the dvgrab main program

    contains the main logic
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// C++ includes

#include <deque>

// C includes

#include <assert.h> 
#include <fcntl.h> 
#include <unistd.h> 
#include <stdio.h> 
#include <stdlib.h> 
#include <signal.h> 
#include <pthread.h> 

// local includes

#include "error.h"
#include "riff.h"
#include "avi.h"
#include "frame.h"
#include "filehandler.h"

#define RAW_BUF_SIZE    (10240)

bool    g_buffer_underrun;
volatile bool    g_reader_active;
volatile int     g_alldone;
char    *g_dst_file_name;

int     g_autosplit;
int     g_timestamp;
int     g_card;
int     g_channel;
int     g_file_format;
bool    g_open_dml;
int     g_frame_count;
int     g_frame_every;
int     g_progress;
int     g_testmode;
int     g_max_file_size;
char   *g_dv1394 = NULL;
FileHandler *g_file = NULL;
int     g_jpeg_quality;
bool    g_jpeg_deinterlace;
int     g_jpeg_width;
int     g_jpeg_height;
bool    g_jpeg_overwrite;

std::deque < Frame* > g_buffer_queue;
std::deque < Frame* > g_output_queue;


pthread_mutex_t g_mutex;
pthread_t       g_thread;

void* read_frames(void* arg);

void usage()
{
    fprintf(stderr, "Usage: dvgrab [options] [file]\n");
    fprintf(stderr, "Try dvgrab --help for more information\n");
}


void getargs(int argc, char *argv[])
{
    int i;

    /* Set up all the defaults */

    g_autosplit = false;
    g_timestamp = false;
    g_card = 0;
    g_channel = 63;
    g_file_format = AVI_DV1_FORMAT;
    g_open_dml = false;
    g_frame_count = 100000;
    g_frame_every = 1;
    g_dst_file_name = NULL;
    g_progress = false;
    g_testmode = false;
	g_max_file_size = 1024;
	g_jpeg_quality = 75;
	g_jpeg_deinterlace = false;
	g_jpeg_width = -1;
	g_jpeg_height = -1;
	g_jpeg_overwrite = false;

    for (i = 1; i < argc; ++i) {
        if (strcmp("--autosplit", argv[i]) == 0) {
            g_autosplit = true;
        } else if (strcmp("--timestamp", argv[i]) == 0) {
            g_timestamp = true;
        } else if ((strcmp("--format", argv[i]) == 0) && (i < argc)) {
            i++;
            if (strcmp("dv1", argv[i]) == 0) {
                g_file_format = AVI_DV1_FORMAT;
            } else if (strcmp("dv2", argv[i]) == 0) {
                g_file_format = AVI_DV2_FORMAT;
            } else if (strcmp("raw", argv[i]) == 0) {
                g_file_format = RAW_FORMAT;
            } else if (strcmp("qt", argv[i]) == 0) {
                g_file_format = QT_FORMAT;
            } else if (strcmp("test", argv[i]) == 0) {
                g_file_format = TEST_FORMAT;
#if defined(HAVE_LIBJPEG) && defined(HAVE_LIBDV)
            } else if (strcmp("jpeg", argv[i]) == 0 || strcmp("jpg", argv[i]) == 0) {
                g_file_format = JPEG_FORMAT;
#endif
            } else {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if ((strcmp("--opendml", argv[i]) == 0)) {
			g_open_dml = true;
        } else if ((strcmp("--size", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_max_file_size) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if ((strcmp("--frames", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_frame_count) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if ((strcmp("--every", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_frame_every) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if ((strcmp("--dv1394", argv[i]) == 0) && (i < argc)) {
            i++;
			g_dv1394 = argv[i];
        } else if ((strcmp("--card", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_card) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if ((strcmp("--channel", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_channel) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if (strcmp("--progress", argv[i]) == 0) {
            g_progress = true;
        } else if (strcmp("--testmode", argv[i]) == 0) {
            g_testmode = true;
#if defined(HAVE_LIBJPEG) && defined(HAVE_LIBDV)
        } else if (strcmp("--jpeg-overwrite", argv[i]) == 0) {
            g_jpeg_overwrite = true;
        } else if ((strcmp("--jpeg-quality", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_jpeg_quality) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if (strcmp("--jpeg-deinterlace", argv[i]) == 0) {
            g_jpeg_deinterlace = true;
        } else if ((strcmp("--jpeg-width", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_jpeg_width) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
        } else if ((strcmp("--jpeg-height", argv[i]) == 0) && (i < argc)) {
            i++;
            if (sscanf(argv[i], "%d", &g_jpeg_height) != 1) {
                usage();
                exit(EXIT_FAILURE);
            }
#endif
        } else if (strcmp("--help", argv[i]) == 0) {
            printf("Usage: %s [options] file\n\n", argv[0]);
			printf("If file is '-', then raw DV is sent to stdout.\n\n" );
            printf("Options:\n\n");
            printf("  --autosplit       start a new file when the frame count is exceeded, a new\n");
            printf("                      recording is detected, or file becomes too big.\n");
            printf("                      Note: will loop forever -- you need to press ctrl-c\n");
            printf("                      at the end.\n");
            printf("                      Default is autosplit when file is close to %d MB.\n", g_max_file_size);
            printf("  --timestamp       put the date and time of recording into the file name.\n");
            printf("  --format dv1      save as a 'Type 1' DV AVI file. %s\n", g_file_format == AVI_DV1_FORMAT ? "This is the default." : " ");
            printf("  --format dv2      save as a 'Type 2' DV AVI file. %s\n", g_file_format == AVI_DV2_FORMAT ? "This is the default." : " ");
            printf("  --opendml         use the OpenDML extensions to write large (>1GB)\n");
			printf("                      'Type 2' DV AVI files (default %s).\n", g_open_dml? "on" : "off");
            printf("  --format raw      save only the raw DV information.\n");
#ifdef HAVE_LIBQUICKTIME
            printf("  --format qt       save as a QuickTime movie.\n");
#endif
#if defined(HAVE_LIBJPEG) && defined(HAVE_LIBDV)
            printf("  --format jpeg     save as a sequence of JPEG files.\n");
            printf("  --jpeg-overwrite  overwrite the same file instead of createing a sequence.\n");
            printf("  --jpeg-quality n  set the JPEG compression level.\n");
			printf("  --jpeg-deinterlace deinterlace the output by line doubling the upper field.\n");
			printf("  --jpeg-width n    scale the output to the specified width (max=2048).\n");
			printf("  --jpeg-height n   scale the output to the specified height (max=2048).\n");
#endif
            printf("  --format test     save in a format suitable as test input for --testmode.\n");
            printf("                      See the documentation for a discussion of the file types.\n");
            printf("  --frames number   number of frames to capture (default %d).\n", g_frame_count);
			printf("  --size number     maximum size of a file, 0 = unlimited (default %d MB).\n", g_max_file_size);
            printf("  --every number    write every n'th frame only (default %d, all frames).\n", g_frame_every);
            printf("  --card number     card number (default %d).\n", g_card);
            printf("  --channel number  iso channel number for listening (default %d).\n", g_channel);
#ifdef HAVE_LIBDV_DV1394_H
            printf("  --dv1394 device   use the dv1394 device named device (default uses raw1394.)\n");
#endif
            //                        printf("  --progress        print a '.' for each received frame.\n");
            printf("  --testmode        read video data from the file 'testdata.raw'\n");
            printf("                      instead from camcorder.\n");
            printf("  --help            display this help and exit.\n");
            printf("  --version         display version information and exit.\n");
            printf("\nCheck out the dvgrab website for the latest version, news and other software:\n");
            printf("http://kino.schirmacher.de/\n\n");
            exit(EXIT_SUCCESS);
        } else if (strcmp("--version", argv[i]) == 0) {
            printf(PACKAGE " " VERSION "\n");
            exit(EXIT_SUCCESS);
        } else if (g_dst_file_name == NULL) {
            g_dst_file_name = argv[i];
			if ( strcmp( g_dst_file_name, "-" ) == 0 )
				g_file_format = RAW_FORMAT;
        } else {
            usage();
            exit(EXIT_FAILURE);
        }
    }
    if (g_dst_file_name == NULL) {
        usage();
        exit(EXIT_FAILURE);
    }
}


Frame* GetNextFrame()
{
    Frame       *frame = NULL;

    while (frame == NULL) {

        pthread_mutex_lock(&g_mutex);

        if (g_output_queue.size() > 0) {
            frame = g_output_queue[0];
            g_output_queue.pop_front();
            // printf("writer < out: buffer %d, output %d\n",g_buffer_queue.size(), g_output_queue.size());
            // fflush(stdout);
        }

        pthread_mutex_unlock(&g_mutex);

        if (!g_reader_active)
            break;

        if (frame == NULL)
            usleep (100000);
    }
	if (frame != NULL)
		frame->ExtractHeader();
    return frame;
}


void DoneWithFrame(Frame *frame)
{
    pthread_mutex_lock(&g_mutex);
    g_buffer_queue.push_back(frame);
    // printf("writer > buf: buffer %d, output %d\n",g_buffer_queue.size(), g_output_queue.size());
    // fflush(stdout);
    pthread_mutex_unlock(&g_mutex);
}


int capture()
{
    int incomplete_frames = 0;
    Frame *frame = NULL;

    g_alldone = false;
    g_buffer_underrun = false;

    for (int i = 0; i < 100; ++i) {
        fail_null(frame = new Frame);
        g_buffer_queue.push_back(frame);
    }
    pthread_mutex_init(&g_mutex, NULL);
    g_reader_active = true;
    pthread_create(&g_thread, NULL, read_frames, NULL);

    while (g_reader_active == true) {
        frame = GetNextFrame();
        if (frame == NULL)
            break;
        if (frame->IsComplete()) {
            break;
        } else {
            DoneWithFrame(frame);
        }
    }

	switch (g_file_format) {
		case RAW_FORMAT:
			g_file = new RawHandler();
			break;
		
		case AVI_DV1_FORMAT:
			g_file = new AVIHandler(AVI_DV1_FORMAT);
			break;
	
		case AVI_DV2_FORMAT:
		{
			AVIHandler *aviWriter = new AVIHandler(AVI_DV2_FORMAT);
			g_file = aviWriter;
			aviWriter->SetOpenDML(g_open_dml);
			break;
		}

#ifdef HAVE_LIBQUICKTIME		
		case QT_FORMAT:
			g_file = new QtHandler();
			break;
#endif

#if defined(HAVE_LIBJPEG) && defined(HAVE_LIBDV)
		case JPEG_FORMAT:
			g_file = new JPEGHandler( g_jpeg_quality, g_jpeg_deinterlace, g_jpeg_width, g_jpeg_height, g_jpeg_overwrite );
			break;
#endif
	}
	g_file->SetTimeStamp(g_timestamp);
	g_file->SetBaseName(g_dst_file_name);
	g_file->SetMaxFrameCount(g_frame_count);
	g_file->SetAutoSplit(g_autosplit);
	g_file->SetEveryNthFrame(g_frame_every);
	g_file->SetMaxFileSize( (off_t) g_max_file_size * (off_t)(1024*1024) );
	g_file->SetSampleFrame(*frame);
	
    while (g_alldone == false) {

		/* report a buffer underrun.  */

		if (g_buffer_underrun == true) {

			TimeCode timeCode;
			struct tm recDate;

			frame->GetTimeCode(timeCode);
			frame->GetRecordingDate(recDate);
			fprintf( stderr, "%s: buffer underrun near: timecode %2.2d:%2.2d:%2.2d.%2.2d date %4.4d.%2.2d.%2.2d %2.2d:%2.2d:%2.2d\n", 
				   g_file->GetFileName().c_str(), timeCode.hour, timeCode.min, timeCode.sec, timeCode.frame,
				   recDate.tm_year + 1900, recDate.tm_mon + 1, recDate.tm_mday,
				   recDate.tm_hour, recDate.tm_min, recDate.tm_sec);
			fprintf( stderr, "This error means that the frames could not be written fast enough.\n"); 
			g_buffer_underrun = false;
		}

		/* report a broken frame. */

		if (frame->IsComplete() == false) {

			TimeCode timeCode;
			struct tm recDate;

			frame->GetTimeCode(timeCode);
			frame->GetRecordingDate(recDate);
			fprintf( stderr, "%s: frame dropped: timecode %2.2d:%2.2d:%2.2d.%2.2d date %4.4d.%2.2d.%2.2d %2.2d:%2.2d:%2.2d\n", 
				   g_file->GetFileName().c_str(), timeCode.hour, timeCode.min, timeCode.sec, timeCode.frame,
				   recDate.tm_year + 1900, recDate.tm_mon + 1, recDate.tm_mday,
				   recDate.tm_hour, recDate.tm_min, recDate.tm_sec);
			fprintf( stderr, "This error means that the ieee1394 driver received an incomplete frame.\n"); 
			incomplete_frames++;
		} else {
			g_file->WriteFrame(*frame);
		}

		DoneWithFrame(frame);

		/* If we have any frames in our output queue, continue processing.
		   If the queue is empty but the receiver thread is still running we expect the queue to fill */

		if ((g_output_queue.size() > 0) || (g_reader_active == true)) {

			frame = GetNextFrame();
			if (frame == NULL)
				break;

			/* If the user does not want autosplit, but has set a maximum frame count, turn everything off
			   and quit grabbing when this count is reached. */

			if (g_autosplit == false && g_file->GetFramesWritten() >= g_frame_count) {
				g_alldone = true;
				g_reader_active = false;
			}
		}
		
		if ( !g_file->FileIsOpen() )
		{
			fprintf( stderr, "%s: %d frames\n", g_file->GetFileName().c_str(), g_file->GetFramesWritten());
			if (incomplete_frames > 0)
				fprintf( stderr, "%d incomplete frames detected.\n", incomplete_frames);
			fflush(stdout);
		}

    }

	if ( g_file->FileIsOpen() )
	{
		fprintf( stderr, "%s: %d frames\n", g_file->GetFileName().c_str(), g_file->GetFramesWritten());
		if (incomplete_frames > 0)
			fprintf( stderr, "%d incomplete frames detected.\n", incomplete_frames);
		fflush(stdout);
		g_file->Close();
	}
	
	delete g_file;
	g_file = NULL;
	
    return 0;
}


extern int capture_test();

void signal_handler(int sig)
{
    /* replace this signal handler with the default (which aborts) */

    signal(SIGINT, SIG_IGN);
    signal(SIGHUP, SIG_IGN);

    /* setting these variables will let us fall out of the main loop */

    g_reader_active = false;
    g_alldone = true;
}


int main(int argc, char *argv[])
{
    try {
        getargs(argc, argv);

        signal(SIGINT, signal_handler);
        signal(SIGHUP, signal_handler);

        switch (g_file_format) {

        case TEST_FORMAT:
            capture_test();
            break;

        case RAW_FORMAT:
        case AVI_DV1_FORMAT:
        case AVI_DV2_FORMAT:
		case QT_FORMAT:
		case JPEG_FORMAT:
            capture();
            break;

        default:
            assert(0);
            break;
        }
    } catch (char *s) {
        fprintf( stderr, "Exception caught.\n%s\n", s);
    }

    catch (...) {
        fprintf( stderr, "Unexpected exception caught.\n");
    }
    return 0;
}

